
#include <iostream>
#include <vector>
#include <stdlib.h>
#include <string>
using namespace std;


bool Cyclic = false;


int h, w;
vector<vector<int> > ChessBoard;
int OffsetX[] = { 2, 1,-1,-2,-2,-1, 1, 2};
int OffsetY[] = {-1,-2,-2,-1, 1, 2, 2, 1};


bool OutOfBounds(int x, int y)
{
    return (x < 0) || (x >= w) || (y < 0) || (y >= h);
}


void Print()
{
    for(int y = 0; y < h; y++)
    {
        for(int x = 0; x < w; x++)
            cout << (ChessBoard[y][x] < 10 ? "   " : (ChessBoard[y][x] < 100 ? "  " : " ")) << ChessBoard[y][x];
        cout << "\n";
    }
    cout << "\033[" << h << "A";
}


void Print(int x, int y)
{
    cout << "\033[" << (y+1) << ";" << (4*x+1) << "H"
         << (ChessBoard[y][x] < 10 ? "   " : (ChessBoard[y][x] < 100 ? "  " : " ")) << ChessBoard[y][x];
}


int WarnsdorffRule(int x, int y)
{
    if(OutOfBounds(x,y) || ChessBoard[y][x] != 0)
        return -1;
    
    int result = 0;
    for(int i = 0; i < 8; i++)
        if(!OutOfBounds(x+OffsetX[i], y+OffsetY[i])
        && ChessBoard[y+OffsetY[i]][x+OffsetX[i]] == 0)
            result++;
    return result;
}


vector<vector<int> > DijkstraStorage;
vector<int> DijkstraQueue;
bool PathsExist(int fromx, int fromy, int tox, int toy, int depth)
{
    vector<int>::iterator queuehead = DijkstraQueue.begin(), queuetail = DijkstraQueue.begin();;
    *queuetail = fromx; queuetail++;
    *queuetail = fromy; queuetail++;
    for(int i = 0; i < h; i++)
        std::fill(DijkstraStorage[i].begin(), DijkstraStorage[i].end(), 0);
    DijkstraStorage[fromy][fromx] = 1;
    int reachable = 1;
    
    while(queuehead < queuetail)
    {
        int x = *queuehead; queuehead++;
        int y = *queuehead; queuehead++;
        
        for(int i = 0; i < 8; i++)
        {
            if(!OutOfBounds(x+OffsetX[i], y+OffsetY[i])
            && ChessBoard[y+OffsetY[i]][x+OffsetX[i]] == 0
            && DijkstraStorage[y+OffsetY[i]][x+OffsetX[i]] == 0)
            {
                reachable++;
                DijkstraStorage[y+OffsetY[i]][x+OffsetX[i]] = 1;
                *queuetail = x+OffsetX[i]; queuetail++;
                *queuetail = y+OffsetY[i]; queuetail++;
            }
        }
    }

    
    if(reachable + depth < w*h - Cyclic)
        return false; // there are unreachable vertices in the residual graph

    
    if(Cyclic)
    {
        // for a cyclic solution, there also needs to be a path back to the start
        for(int i = 0; i < 8; i++)
            if(!OutOfBounds(tox+OffsetX[i], toy+OffsetY[i]))
                if(DijkstraStorage[toy+OffsetY[i]][tox+OffsetX[i]] == 1)
                    return true;
        return false;
    }
    else
        return true;
}


bool KnightsTour(int x, int y, int depth)
{
    if(OutOfBounds(x,y))
        return false;
    
    if(Cyclic)                                      // acceptance condition for cycles
        if(depth == h*w && ChessBoard[y][x] == h*w) 
            return true;

    if(ChessBoard[y][x] != 0)
        return false;

    if(!PathsExist(x,y, 1,2, depth))
        return false;
    
    ChessBoard[y][x] = depth;
    Print(x,y);

    if(!Cyclic)                                    // acceptance condition for paths
        if(depth == h*w)
            return true;
    
    // selection of order in which neighbors are tried (using Warnsdorff Rule)
    int WarnsdorffIndex[8];
    int WarnsdorffValue[8];
    for(int i = 0; i < 8; i++)
    {
        WarnsdorffIndex[i] = i;
        WarnsdorffValue[i] = WarnsdorffRule(x+OffsetX[i], y+OffsetY[i]);
    }
    
    // selection sort (best on small arrays like these)
    for(int i = 0; i < 8; i++)
    {
        int bestj = i;
        for(int j = i+1; j < 8; j++)
            if(WarnsdorffValue[j] < WarnsdorffValue[bestj])
                bestj = j;
        int temp = WarnsdorffValue[i]; WarnsdorffValue[i] = WarnsdorffValue[bestj]; WarnsdorffValue[bestj] = temp;
            temp = WarnsdorffIndex[i]; WarnsdorffIndex[i] = WarnsdorffIndex[bestj]; WarnsdorffIndex[bestj] = temp;
    }
    
    // recursion
    for(int i = 0; i < 8; i++)
        if(KnightsTour(x+OffsetX[WarnsdorffIndex[i]],y+OffsetY[WarnsdorffIndex[i]],depth+1))
            return true;

    ChessBoard[y][x] = 0;
    Print(x,y);
    return false;
}


int main(int argc, char** argv)
{
    cout << "\033[2J\033[1;1H";
    
    w = 8, h = 8;                      // set height and width from parameters
    bool firstnum = true;
    for(int i = 1; i < argc; i++)
    {
        if(string(argv[i]) == "--cyclic")
            Cyclic = true;
        else
        {
            int temp = atoi(argv[i]);
            if(firstnum)
                h = temp;
            w = temp;
        }
    }

    
    ChessBoard = vector<vector<int> >(h, vector<int>(w, 0));  // allocate memory
    DijkstraStorage = vector<vector<int> >(h, vector<int>(w, 0));  // allocate memory
    DijkstraQueue = vector<int>(2*w*h+1, 0);
    if(Cyclic)
    {
        ChessBoard[0][0] = 1; // can be assumed w.l.o.g.
        ChessBoard[1][2] = 2;
        ChessBoard[2][1] = w*h;
    }
    Print();
    

    bool found = false;
    if(Cyclic)
    {
        // w.l.o.g. a cycle starts at (0,0), visits (2,1) next and returns through (1,2)
        for(int i = 0; i < 8 && !found; i++)
            if(KnightsTour(2+OffsetX[i],1+OffsetY[i],3))
                found = true;
    }
    else
    {
        // a path could start anywhere (only avoid symmetric start choices)
        for(int i = (h-1)/2; i >= 0 && !found; --i)
            for(int j = (w-1)/2; j >= (w==h?i:0) && !found; --j)
                if(KnightsTour(j,i,1))
                    found = true;
    }
    
    if(found)
    {
        cout << "\033[1;1H";
        Print();
        cout << "\033[" << (h+1) << ";1H";
    }
    else
        cout << "\033[2J\033[1;1HNo solution!\n";
    return 0;
}
